import os
import time
import socket
from six.moves import BaseHTTPServer
from six.moves.SimpleHTTPServer import SimpleHTTPRequestHandler
from multiprocessing import Process, Event


def start_server():
    port = free_port()
    server_address = ('127.0.0.1', port)
    root = os.path.join(os.path.dirname(__file__), 'www')
    initialized = Event()
    process = Process(
        target=run_server,
        args=(root, server_address, initialized),
    )
    process.start()
    initialized.wait()
    time.sleep(0.1)
    return process, server_address


def stop_server(process):
    process.terminate()


def run_server(root, server_address, initialized):
    os.chdir(root)
    SimpleHTTPRequestHandler.protocol_version = "HTTP/1.0"
    httpd = BaseHTTPServer.HTTPServer(server_address, SimpleHTTPRequestHandler)
    sa = httpd.socket.getsockname()
    print("Serving HTTP on {0} port {1}...".format(sa[0], sa[1]))
    initialized.set()
    httpd.serve_forever()


def free_port():
    ports_to_check = [17261, 0]
    for port in ports_to_check:
        s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        try:
            s.bind(('', port))
        except (OSError, socket.error):
            # Address already in use
            continue
        addr = s.getsockname()
        s.close()
        return addr[1]



import logging

import requests

from ..conf import settings


logger = logging.getLogger(__name__)


class MailgunNotify(object):

    def __init__(self, conf, value):
        mailgun_creds = settings().creds.get('mailgun', {})
        mailgun_creds.update(value or {})
        domain = mailgun_creds['domain']
        self.context = {
            'subject': 'Kibitzr update for ' + conf['name'],
            'from': 'Kibitzer <mailgun@{0}>'.format(domain),
            'to': [mailgun_creds['to']],
        }
        self.url = 'https://api.mailgun.net/v3/{0}/messages'.format(domain)
        self.auth = ('api', mailgun_creds['key'])
        self.session = requests.Session()

    def __call__(self, report):
        response = self.session.post(
            self.url,
            auth=self.auth,
            data=self.payload(report),
        )
        logger.debug(response.text)
        response.raise_for_status()
        return response

    def payload(self, report):
        return dict(
            self.context,
            text=report,
        )


def notify_factory(conf, value):
    return MailgunNotify(conf, value)



import logging
import pkgutil
import importlib


logger = logging.getLogger(__name__)


def dummy_notify_factory(notify_func):
    def factory(conf, value):
        return notify_func
    return factory


def load_notifiers():
    path = pkgutil.extend_path(__path__, __name__)
    prefix = __name__ + '.'
    registry = {}
    for importer, modname, ispkg in pkgutil.walk_packages(path, prefix):
        submodule = importlib.import_module(modname, __name__)
        if hasattr(submodule, 'register'):
            submodule.register(registry)
        else:
            key = getattr(submodule, 'NAME', modname.split('.')[-1])
            if hasattr(submodule, 'notify_factory'):
                registry[key] = submodule.notify_factory
            elif hasattr(submodule, 'notify'):
                registry[key] = dummy_notify_factory(submodule.notify)
    return registry


def create_notifier(name, conf, value):
    try:
        notify_factory = REGISTRY[name]
    except KeyError:
        logger.error("Unknown notifier %r", name)
        return None
    else:
        return notify_factory(conf=conf, value=value)


REGISTRY = load_notifiers()



from .browser import firefox_fetcher, cleanup, persistent_firefox  # noqa
from .simple import SessionFetcher  # noqa
from .script import fetch_by_script  # noqa


def cleanup_fetchers():
    cleanup()



from kibitzr.notifier.webhook import WebHookNotify

from ..compat import mock


def test_webhook_sample():
    url = 'http://localhost:0'
    notify = WebHookNotify(value=url)
    with mock.patch.object(notify.session, 'post') as fake_post:
        notify('report')
    fake_post.assert_called_once_with(
        url=url,
        data={'message': 'report'},
    )



import tempfile
import logging
import traceback

import sh
from ..conf import settings


logger = logging.getLogger(__name__)
PYTHON_ERROR = "script.python must set global variables ok and content"


def fetch_by_script(conf, **_kwargs):
    code = conf['script']
    try:
        python_code = code['python']
    except (KeyError, TypeError):
        # Not a python script
        pass
    else:
        return fetch_by_python(python_code, conf)
    try:
        # Explicit notation:
        bash_script = code['bash']
    except (KeyError, TypeError):
        bash_script = code
    return fetch_by_bash(bash_script)


def fetch_by_bash(code):
    logger.info("Executing bash fetcher")
    logger.debug(code)
    with tempfile.NamedTemporaryFile() as fp:
        logger.debug("Saving code to %r", fp.name)
        fp.write(code.encode('utf-8'))
        fp.flush()
        logger.debug("Launching script %r", fp.name)
        result = sh.bash(fp.name)
        stdout = result.stdout.decode('utf-8')
        stderr = result.stderr.decode('utf-8')
        logger.debug("Bash exit_code: %r", result.exit_code)
        logger.debug("Bash stdout: %s", stdout)
        logger.debug("Bash stderr: %s", stderr)
        ok = (result.exit_code == 0)
        if ok:
            report = stdout
        else:
            report = u'\n'.join([stdout, stderr])
        return ok, report


def fetch_by_python(code, conf):
    logger.info("Fetch using Python script")
    logger.debug(code)
    assert 'ok' in code, PYTHON_ERROR
    assert 'content' in code, PYTHON_ERROR
    try:
        # ok, content = False, None
        namespace = {}
        exec(code, {'conf': conf, 'creds': settings().creds}, namespace)
        return namespace['ok'], namespace['content']
    except:
        logger.exception("Python fetcher raised an Exception")
        return False, traceback.format_exc()



import sh
import tempfile
import logging


logger = logging.getLogger(__name__)


class BashNotify(object):

    def __init__(self, conf, value):
        self.code = value

    def __call__(self, report):
        logger.info("Executing custom notifier")
        logger.debug(self.code)
        with tempfile.NamedTemporaryFile() as fp:
            logger.debug("Saving code to %r", fp.name)
            fp.write(self.code.encode('utf-8'))
            fp.flush()
            logger.debug("Launching script %r", fp.name)
            result = sh.bash(fp.name, _in=report)
            logger.debug("Bash exit_code: %r", result.exit_code)
            logger.debug("Bash stdout: %s", result.stdout.decode('utf-8'))
            logger.debug("Bash stderr: %s", result.stderr.decode('utf-8'))


def notify_factory(conf, value):
    return BashNotify(conf, value)



import logging
import smtplib

import six

from ..conf import settings


logger = logging.getLogger(__name__)


def notify(conf, report, notifier_conf, **kwargs):
    logger.info("Executing SMTP notifier")
    credentials = settings().creds['smtp']
    user = credentials['user']
    password = credentials['password']
    host = credentials['host']
    port = credentials['port']
    try:
        recipients = notifier_conf['recipients']
    except (TypeError, KeyError) as exc:
        recipients = notifier_conf
    if isinstance(recipients, six.string_types):
        recipients = [recipients]
    try:
        subject = notifier_conf['subject']
    except (TypeError, KeyError) as exc:
        subject = "Kibitzr update for " + conf['name']
    send_email(
        user=user,
        password=password,
        recipients=recipients,
        subject=subject,
        body=report,
        host=host,
        port=port,
    )


def send_email(user, password, recipients, subject, body, host, port):
    message = (
        u"From: {user}\r\n"
        u"To: {to}\r\n"
        u"Subject: {subject}\r\n\r\n"
        u"{body}\r\n"
        .format(
            user=user,
            to=", ".join(recipients),
            subject=subject,
            body=body,
        )
    )
    try:
        server = smtplib.SMTP(host, port)  # ("smtp.gmail.com", 587)
        server.ehlo()
        server.starttls()
        server.login(user, password)
        server.sendmail(user, recipients, message.encode("utf-8"))
        server.close()
        logger.debug('Successfully sent the mail')
    except:
        logger.exception("Failed to send the e-mail")



"""
Built-in transforms
"""
import sys
import logging
import contextlib
import functools
import json
from lxml import etree
import traceback

import six
from bs4 import BeautifulSoup
import sh
from lazy_object_proxy import Proxy as Lazy

from .storage import PageHistory
from .conf import settings


PYTHON_ERROR = "transform.python must set global variables ok and content"
logger = logging.getLogger(__name__)
jq = Lazy(lambda: sh.jq.bake('--monochrome-output', '--raw-output'))


def pipeline_factory(conf):
    rules = conf.get('transform', [])
    if isinstance(rules, six.string_types):
        rules = [rules]
    return functools.partial(
        pipeline,
        transformers=[
            transformer_factory(conf, rule)
            for rule in rules
        ]
    )


def pipeline(ok, content, transformers):
    for transformer in transformers:
        if ok:
            ok, content = transformer(content)
        else:
            break
    return ok, content


def transformer_factory(conf, rule):
    try:
        name, value = next(iter(rule.items()))
    except AttributeError:
        name, value = rule, None
    if name == 'css':
        return functools.partial(css_selector, value)
    if name == 'css-all':
        return functools.partial(css_selector, value, select_all=True)
    elif name == 'xpath':
        return functools.partial(xpath_selector, value)
    elif name == 'tag':
        return functools.partial(tag_selector, value)
    elif name == 'text':
        return extract_text
    elif name == 'changes':
        if value and value.lower() == 'verbose':
            return functools.partial(PageHistory(conf).report_changes,
                                     verbose=True)
        else:
            return PageHistory(conf).report_changes
    elif name == 'json':
        return pretty_json
    elif name == 'jq':
        return functools.partial(run_jq, value)
    elif name == 'sort':
        return sort_lines
    elif name == 'cut':
        return functools.partial(cut_lines, value)
    elif name == 'python':
        return functools.partial(python_transform, conf=conf, code=value)
    else:
        raise RuntimeError(
            "Unknown transformer: %r" % (name,)
        )


def pretty_json(text):
    json_dump = json.dumps(
        json.loads(text),
        indent=2,
        sort_keys=True,
        ensure_ascii=False,
        # encoding='utf-8',
    )
    return True, u'\n'.join([
        line.rstrip()
        for line in json_dump.splitlines()
    ])


def tag_selector(name, html):
    with deep_recursion():
        soup = BeautifulSoup(html, "html.parser")
        element = soup.find(name)
        if element:
            return True, six.text_type(element)
        else:
            logger.warning('Tag not found: %r', name)
            return False, html


def css_selector(selector, html, select_all=False):
    with deep_recursion():
        soup = BeautifulSoup(html, "html.parser")
        try:
            elements = soup.select(selector)
            if select_all:
                result = u"".join(six.text_type(x)
                                  for x in elements)
            else:
                result = six.text_type(elements[0])
            return True, result
        except IndexError:
            logger.warning('CSS selector not found: %r', selector)
            return False, html


def xpath_selector(selector, html):
    root = etree.fromstring(html, parser=etree.HTMLParser())
    elements = root.xpath(selector)
    if elements:
        return True, etree.tostring(
            next(iter(elements)),
            method='html',
            pretty_print=True,
            encoding='unicode',
        )
    else:
        logger.warning('XPath selector not found: %r', selector)
        return False, html


def extract_text(html):
    with deep_recursion():
        strings = BeautifulSoup(html, "html.parser").stripped_strings
        return True, u'\n'.join([
            line
            for line in strings
            if line
        ])


def sort_lines(text):
    return True, u''.join([
        line + u'\n'
        for line in sorted(text.splitlines())
        if line
    ])


def cut_lines(last_line, text):
    return True, u''.join([
        line + u'\n'
        for line in text.splitlines()[:last_line]
    ])


def run_jq(query, text):
    logger.debug("Running jq query %s against %s", query, text)
    try:
        command = jq(query, _in=text)
        if not command.stderr:
            success, result = True, command.stdout.decode('utf-8')
        else:
            success, result = False, command.stderr.decode('utf-8')
    except sh.ErrorReturnCode as exc:
        logger.exception("jq failure")
        success, result = False, exc.stderr
    logger.debug("jq transform success: %r, content: %r",
                 success, result)
    return success, result


def python_transform(content, code, conf):
    logger.info("Python transform")
    logger.debug(code)
    assert 'ok' in code, PYTHON_ERROR
    assert 'content' in code, PYTHON_ERROR
    try:
        namespace = {'content': content}
        exec(code, {'creds': settings().creds, 'conf': conf}, namespace)
        return namespace['ok'], six.text_type(namespace['content'])
    except:
        logger.exception("Python transform raised an Exception")
        return False, traceback.format_exc()


@contextlib.contextmanager
def deep_recursion():
    old_limit = sys.getrecursionlimit()
    try:
        sys.setrecursionlimit(100000)
        yield
    finally:
        sys.setrecursionlimit(old_limit)



import logging

import requests

from ..conf import settings


logger = logging.getLogger(__name__)


class WebHookNotify(object):

    CREDS_KEY = 'webhook'
    POST_KEY = 'message'

    def __init__(self, creds_key=None, value=None, post_key=None):
        self.session = requests.Session()
        self.url = self.load_url(creds_key or self.CREDS_KEY, value)
        self.post_key = post_key or self.POST_KEY
        self.configure_session()

    @staticmethod
    def load_url(creds_key, value):
        if value:
            return value
        else:
            webhook_creds = settings().creds[creds_key]
            return webhook_creds['url']

    def post(self, report, **kwargs):
        response = self.session.post(
            url=self.url,
            data=self.payload(report),
        )
        logger.debug(response.text)
        response.raise_for_status()
        return response

    def configure_session(self):
        pass

    def payload(self, report):
        return {self.post_key: report}

    __call__ = post


def webhook_factory(klass):
    def notify_factory(conf, value):
        return klass(value=value)
    return notify_factory



import os
import sys
import time
import shutil
import logging
from contextlib import contextmanager

from xvfbwrapper import Xvfb
from selenium import webdriver
from selenium.webdriver.firefox.firefox_binary import FirefoxBinary
from selenium.common.exceptions import (
    WebDriverException,
    StaleElementReferenceException,
)

from ..conf import settings


logger = logging.getLogger(__name__)
PROFILE_DIR = 'firefox_profile'
HOME_PAGE = 'https://kibitzr.github.io/'


firefox_instance = {
    'xvfb_display': None,
    'driver': None,
    'headed_driver': None,
}


def cleanup():
    """Must be called before exit"""
    global firefox_instance
    temp_dirs = []
    if firefox_instance['driver'] is not None:
        if firefox_instance['driver'].profile:
            temp_dirs.append(firefox_instance['driver'].profile.profile_dir)
        firefox_instance['driver'].quit()
        firefox_instance['driver'] = None
    if firefox_instance['headed_driver'] is not None:
        if firefox_instance['headed_driver'].profile:
            temp_dirs.append(firefox_instance['headed_driver'].profile.profile_dir)
        firefox_instance['headed_driver'].quit()
        firefox_instance['headed_driver'] = None
    if firefox_instance['xvfb_display'] is not None:
        firefox_instance['xvfb_display'].stop()
        firefox_instance['xvfb_display'] = None
    for temp_dir in temp_dirs:
        shutil.rmtree(temp_dir, ignore_errors=True)


def persistent_firefox():
    if not os.path.exists(PROFILE_DIR):
        os.makedirs(PROFILE_DIR)
    with firefox(headless=False) as driver:
        driver.get(HOME_PAGE)
        while True:
            try:
                # Property raises when browser is closed:
                driver.title
            except:
                # All kinds of things happen when closing Firefox
                break
            else:
                time.sleep(0.2)
        update_profile(driver)


def update_profile(driver):
    if os.path.exists(PROFILE_DIR):
        shutil.rmtree(PROFILE_DIR)
    shutil.copytree(
        driver.profile.profile_dir,
        PROFILE_DIR,
        ignore=shutil.ignore_patterns(
            "parent.lock",
            "lock",
            ".parentlock",
            "*.sqlite-shm",
            "*.sqlite-wal",
        ),
    )


def firefox_fetcher(conf):
    url = conf['url']
    delay = conf.get('delay')
    scenario = conf.get('scenario')
    headless = conf.get('headless', True)
    with firefox(headless) as driver:
        driver.get(url)
        if scenario:
            run_scenario(driver, scenario, conf)
        if delay:
            time.sleep(delay)
        elem = driver.find_element_by_xpath("//*")
        for attempt in range(3):
            try:
                html = elem.get_attribute("outerHTML")
            except StaleElementReferenceException:
                # Crazy (but stable) race condition,
                # new page loaded after call to find_element_by_xpath
                # Just retry:
                time.sleep(1)
                elem = driver.find_element_by_xpath("//*")
            else:
                break
        # Create a new tab and close the old one
        # to avoid idle page resource usage
        old_tab = driver.current_window_handle
        driver.execute_script('''window.open("about:blank", "_blank");''')
        driver.switch_to.window(old_tab)
        driver.close()
        driver.switch_to.window(driver.window_handles[0])
        return True, html


@contextmanager
def firefox(headless=True):
    global firefox_instance
    if headless:
        if firefox_instance['xvfb_display'] is None:
            firefox_instance['xvfb_display'] = virtual_buffer()
        driver_key = 'driver'
    else:
        driver_key = 'headed_driver'
    if firefox_instance[driver_key] is None:
        if logger.level == logging.DEBUG:
            firefox_binary = FirefoxBinary(log_file=sys.stdout)
        else:
            firefox_binary = None
        # Load profile, if it exists:
        if os.path.isdir(PROFILE_DIR):
            firefox_profile = webdriver.FirefoxProfile(PROFILE_DIR)
        else:
            firefox_profile = None
        firefox_instance[driver_key] = webdriver.Firefox(
            firefox_binary=firefox_binary,
            firefox_profile=firefox_profile,
        )
        firefox_instance[driver_key].set_window_size(1024, 768)
    yield firefox_instance[driver_key]


def virtual_buffer():
    """
    Try to start xvfb, trying multiple (up to 5) times if a failure
    """
    for _ in range(0, 6):
        xvfb_display = Xvfb()
        xvfb_display.start()
        # If Xvfb started, return.
        if xvfb_display.proc is not None:
            return xvfb_display
    raise Exception("Xvfb could not be started after six attempts.")


def run_scenario(driver, code, conf):
    logger.info("Executing custom scenario")
    logger.debug(code)
    exec(
        code,
        {
            'conf': conf,
            'creds': settings().creds,
            'driver': driver,
        },
    )



import pytest
from kibitzr import main
from ..compat import mock
from ..helpers import SettingsMock


@pytest.fixture(autouse=True)
def settings():
    """Override native settings singleton with empty one"""
    SettingsMock.instance()


@pytest.fixture(autouse=True)
def check_noop(mocker):
    """Cancel all side-effects from checks"""
    mocker.patch("kibitzr.main.Checker.check", spec=True)


def test_loop_aborts_without_checks():
    assert 1 == main.main()


@mock.patch.object(main, "check_forever", side_effect=main.on_interrupt)
def test_main_executes_all_checks_before_loop(the_loop):
    main.settings().pages.append({
        'name': 'A',
        'script': {'python': 'ok, content = True, "ok"'}
    })
    assert 0 == main.main()
    assert the_loop.call_count == 1
    assert the_loop.call_args[0][0][0].check.call_count == 1


def test_dummy_schedule():
    main.Checker.check.side_effect = interrupt_on_nth_call(2)
    main.settings().pages.append({
        'name': 'A',
        'script': {'python': 'ok, content = True, "ok"'},
        'period': 0,
    })
    assert 0 == main.main()
    assert main.Checker.check.call_count == 2


def interrupt_on_nth_call(n):
    def interrupter(*args, **kwargs):
        if interrupter.n > 1:
            interrupter.n -= 1
        else:
            main.on_interrupt()
        return None
    interrupter.n = n
    return interrupter



import sys
import logging

import click

from kibitzr import __version__ as kibitzr_version
from kibitzr.main import main, run_firefox


LOG_LEVEL_CODES = {
    "debug": logging.DEBUG,
    "info": logging.INFO,
    "warning": logging.WARNING,
    "error": logging.ERROR,
}


@click.command()
@click.option("--once", is_flag=True,
              help="Run checks once and exit")
@click.option("-l", "--log-level", default="info",
              type=click.Choice(LOG_LEVEL_CODES.keys()),
              help="Logging level")
@click.option("-v", "--version", is_flag=True,
              help="Print version and exit")
@click.option("--firefox", is_flag=True,
              help="Launch Firefox with persistent profile and exit")
@click.argument('name', nargs=-1)
def entry(once, log_level, version, firefox, name):
    """Run kibitzr in the foreground mode"""
    if version:
        print(kibitzr_version)
    elif firefox:
        run_firefox()
    else:
        log_level_code = LOG_LEVEL_CODES[log_level]
        sys.exit(main(once=once, log_level=log_level_code, names=name))


if __name__ == "__main__":
    entry()



import json

from .webhook import WebHookNotify, webhook_factory


class SlackNotify(WebHookNotify):
    CREDS_KEY = 'slack'

    def payload(self, report):
        return json.dumps({"text": report})


notify_factory = webhook_factory(SlackNotify)



import os
import re
import copy
import logging.config
import contextlib

import yaml


logger = logging.getLogger(__name__)


class ConfigurationError(RuntimeError):
    pass


class ReloadableSettings(object):
    _instance = None
    CONFIG_DIRS = (
        '',
        '~/.config/kibitzr/',
        '~/',
    )
    CONFIG_FILENAME = 'kibitzr.yml'
    CREDENTIALS_FILENAME = 'kibitzr-creds.yml'
    RE_PUNCTUATION = re.compile(r'\W+')
    UNNAMED_PATTERN = 'Unnamed check {0}'

    def __init__(self, config_dir):
        self.filename = os.path.join(config_dir, self.CONFIG_FILENAME)
        self.creds_filename = os.path.join(config_dir,
                                           self.CREDENTIALS_FILENAME)
        self.pages = None
        self.notifiers = None
        self.creds = {}
        self.reread()

    @classmethod
    def detect_config_dir(cls):
        candidates = [
            (directory, os.path.join(directory, cls.CONFIG_FILENAME))
            for directory in map(os.path.expanduser, cls.CONFIG_DIRS)
        ]
        for directory, file_path in candidates:
            if os.path.exists(file_path):
                return directory
        raise ConfigurationError(
            "kibitzr.yml not found in following locations: %s"
            % ", ".join([x[1] for x in candidates])
        )

    @classmethod
    def instance(cls):
        if cls._instance is None:
            config_dir = cls.detect_config_dir()
            cls._instance = cls(config_dir)
        return cls._instance

    def reread(self):
        """
        Read configuration file and substitute references into pages conf
        """
        logger.debug("Loading settings from %s",
                     os.path.abspath(self.filename))
        conf = self.read_conf()
        changed = self.read_creds()
        pages = conf.get('checks', conf.get('pages', []))
        notifiers = conf.get('notifiers', {})
        templates = conf.get('templates', {})
        scenarios = conf.get('scenarios', {})
        unnamed_check_counter = 1
        pages = list(self.unpack_batches(pages))
        for i, page in enumerate(pages):
            if not page.get('name'):
                if page.get('url'):
                    page['name'] = self.url_to_name(page['url'])
                else:
                    page['name'] = self.UNNAMED_PATTERN.format(unnamed_check_counter)
                    unnamed_check_counter += 1
            name = page['name']
            if 'template' in page:
                if page['template'] in templates:
                    templated_page = copy.deepcopy(templates[page['template']])
                else:
                    raise ConfigurationError(
                        "Template %r not found. Referenced in page %r"
                        % (page['template'], name)
                    )
                templated_page.update(page)
                page = templated_page
                del page['template']
                pages[i] = page
            if 'scenario' in page:
                if page['scenario'] in scenarios:
                    page['scenario'] = scenarios[page['scenario']]
            if 'notify' in page:
                for notify in page['notify']:
                    if hasattr(notify, 'keys'):
                        notify_type = next(iter(notify.keys()))
                        notify_param = next(iter(notify.values()))
                        try:
                            notify[notify_type] = notifiers[notify_param]
                        except (TypeError, KeyError):
                            # notify_param is not a predefined notifier name
                            # Save it as is:
                            notify[notify_type] = notify_param
        if self.pages != pages or self.notifiers != notifiers:
            self.pages = pages
            self.notifiers = notifiers
            return True
        else:
            return changed

    @classmethod
    def url_to_name(cls, url):
        return cls.RE_PUNCTUATION.sub('-', url)

    def unpack_batches(self, pages):
        for page in pages:
            if 'batch' in page:
                base = copy.deepcopy(page)
                batch = base.pop('batch')
                url_pattern = base.pop('url-pattern')
                items = base.pop('items')
                for item in items:
                    new_page = copy.deepcopy(base)
                    new_page['name'] = batch.format(item)
                    new_page['url'] = url_pattern.format(item)
                    yield new_page
            else:
                yield page

    def read_conf(self):
        """
        Read and parse configuration file
        """
        with self.open_conf() as fp:
            return yaml.load(fp)

    @contextlib.contextmanager
    def open_conf(self):
        with open(self.filename) as fp:
            yield fp

    def read_creds(self):
        """
        Read and parse credentials file.
        If something goes wrong, log exception and continue.
        """
        logger.debug("Loading credentials from %s",
                     os.path.abspath(self.creds_filename))
        creds = {}
        try:
            with self.open_creds() as fp:
                creds = yaml.load(fp)
        except IOError:
            logger.info("No credentials file found at %s",
                        os.path.abspath(self.creds_filename))
        except:
            logger.exception("Error loading credentials file")
        if creds != self.creds:
            self.creds = creds
            return True
        return False

    @contextlib.contextmanager
    def open_creds(self):
        with open(self.creds_filename) as fp:
            yield fp


def settings():
    """
    Returns singleton instance of settings
    """
    return ReloadableSettings.instance()


logging.config.dictConfig({
    'version': 1,
    'disable_existing_loggers': False,
    'formatters': {
        'standard': {
            'format': '%(asctime)s [%(levelname)s] %(name)s: %(message)s'
        },
    },
    'handlers': {
        'default': {
            'level': 'DEBUG',
            'class': 'logging.StreamHandler',
            'formatter': 'standard',
        },
    },
    'loggers': {
        '': {
            'handlers': ['default'],
            'level': 'INFO',
            'propagate': True,
        },
        'sh': {
            'level': 'INFO',
        },
        'sh.command': {
            'level': 'WARNING',
        },
    }
})



import pytest
from kibitzr.notifier import smtp

from ..compat import mock
from ..helpers import SettingsMock


@pytest.fixture()
def settings():
    """Override native settings singleton with empty one"""
    return SettingsMock.instance()


@mock.patch.object(smtp, 'send_email')
def test_smtp_shortcut(fake_send_email, settings):
    settings.creds.update({
        'smtp': {
            'user': 'user',
            'password': 'password',
            'host': 'host',
            'port': 'port',
        }
    })
    smtp.notify({'name': 'Name'}, 'report', 'you@site.com')
    fake_send_email.assert_called_once_with(
        user='user',
        password='password',
        recipients=['you@site.com'],
        subject='Kibitzr update for Name',
        body='report',
        host='host',
        port='port',
    )


@mock.patch.object(smtp, 'send_email')
def test_smtp_explicit_form(fake_send_email, settings):
    settings.creds.update({
        'smtp': {
            'user': 'user',
            'password': 'password',
            'host': 'host',
            'port': 'port',
        }
    })
    smtp.notify(
        {'name': 'Name'},
        'report',
        {'recipients': ['you@site.com'], 'subject': 'subject'},
    )
    fake_send_email.assert_called_once_with(
        user='user',
        password='password',
        recipients=['you@site.com'],
        subject='subject',
        body='report',
        host='host',
        port='port',
    )
